#include "stdlib.h"
#include "stdio.h" 
#include "fxlib.h"

#include "MonochromeLib.h"
#include "key.h"
#include "time.h"
#include "hero.h"
#include "map.h"

#define MAP_DX 1.6
#define CAMERA_SPEED_Y 1.3

typedef struct {
	double x, y;
	double dx, dy;
	int last_update;
} Camera;

unsigned char block[] = {0x10, 0x10, 0x10, 0xFF, 0x04, 0x04, 0x04, 0xFF};
unsigned char block_top[] = {0xFF, 0x10, 0x10, 0xFF, 0x04, 0x04, 0x04, 0xFF};
unsigned char pick[] = {0x00, 0x18, 0x24, 0x24, 0x52, 0x46, 0x91, 0xA1};
unsigned char key[] = {0x0E, 0x11, 0x15, 0x51, 0xA6, 0x48, 0x90, 0x60};
unsigned char door[] = {0x1F, 0x80, 0x28, 0x40, 0x5F, 0xA0, 0x50, 0xA0, 0xA0, 0x50, 0xA0, 0x70, 0xA0, 0x50, 0xE1, 0x50, 0xA0, 0x50, 0xA0, 0x50, 0xA0, 0x70, 0xA0, 0x50};
unsigned char grass[] = {0xFF, 0xFF, 0xFF, 0xEF, 0x66, 0x00, 0x00, 0x00};
unsigned char box[] = {0xFF, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xFF, 0x80};
unsigned char pick_v2[] = {0x00, 0x18, 0x3C, 0x3C, 0x72, 0x62, 0xE9, 0xC5};
unsigned char platform_right[] = {0xFE, 0x01, 0x01, 0x01, 0xFF, 0x40, 0x40, 0x40};
unsigned char platform_left[] = {0x7F, 0x80, 0x80, 0x80, 0xFF, 0x02, 0x02, 0x02};
unsigned char * map_bmp[] = {block, block_top, pick_v2, key, door, grass, box, platform_left, platform_right};
unsigned char run1[] = {0x5E, 0x00, 0x7F, 0x80, 0x71, 0x80, 0x25, 0x00, 0x45, 0x00, 0x31, 0x80, 0x4E, 0x80, 0x4D, 0x00, 0x33, 0x80, 0x67, 0x80, 0x7B, 0x00, 0x60, 0x00};
unsigned char stop[] = {0x5E, 0x00, 0x7F, 0x80, 0x71, 0x80, 0x25, 0x00, 0x45, 0x00, 0x31, 0x00, 0x2E, 0x00, 0x2D, 0x00, 0x29, 0x00, 0x12, 0x00, 0x1C, 0x00, 0x1E, 0x00};
unsigned char run2[] = {0x5E, 0x00, 0x7F, 0x80, 0x71, 0x80, 0x25, 0x00, 0x45, 0x00, 0x31, 0x00, 0x4E, 0x00, 0x2E, 0x00, 0x61, 0x00, 0x79, 0x00, 0x67, 0x00, 0x07, 0x80};
unsigned char jump[] = {0x5E, 0x00, 0x7F, 0x80, 0x71, 0x80, 0x25, 0x00, 0x45, 0x80, 0xB1, 0x40, 0x4E, 0x80, 0x2D, 0x00, 0x21, 0x00, 0xF3, 0xC0, 0xFF, 0xC0, 0x61, 0x80};
unsigned char * hero_bmp[] = {stop, run1, run2, jump};

unsigned char map[] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
unsigned char map_backup[] = {
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
int map_width = 18;
int map_height = 8;
double map_x = 0, map_y = 0;
double map_dx = 0, map_dy = 0;
int map_last_update;


void map_draw(Camera * camera) {
	int i, j;
	
	int x = camera->x + map_x, y = camera->y + map_y;
	
	for(i=0;i<map_width*map_height;i++) {
		if(x+i%map_width*8 >= -8 && y+i%map_width*8 >= -8 && x+i%map_width*8 <= 123+8 && y+i/map_width*8 <= 63+8) {
			switch(map[i/map_width*map_width+i%map_width]) {
				case BLOCK: ML_bmp_8_or_cl(map_bmp[0], x+i%map_width*8, y+i/map_width*8); break; // Block
				case BLOCK_TOP: ML_bmp_8_or_cl(map_bmp[1], x+i%map_width*8, y+i/map_width*8); break; // Block top
				case PICK: ML_bmp_8_or_cl(map_bmp[2], x+i%map_width*8, y+i/map_width*8); break; // pick
				case KEY: ML_bmp_8_or_cl(map_bmp[3], x+i%map_width*8, y+i/map_width*8); break; // key
				case DOOR: ML_bmp_or_cl(map_bmp[4], x+i%map_width*8, y+i/map_width*8, 12, 12); break; // door
				case BOX: 
					// ML_rectangle(x+i%map_width*8, y+i/map_width*8, x+i%map_width*8+8, y+i/map_width*8+8, 1, ML_BLACK, ML_TRANSPARENT);
					ML_bmp_or_cl(map_bmp[6], x+i%map_width*8, y+i/map_width*8, 9, 9);
				break;
				case GRASS: ML_bmp_8_or_cl(map_bmp[5], x+i%map_width*8, y+i/map_width*8); break; // grass
				case HOLE_LEFT:
					ML_vertical_line(x+i%map_width*8-1, y+i/map_width*8, y+i/map_width*8+8, ML_BLACK);
				break;
				case HOLE_RIGHT:
					ML_vertical_line(x+i%map_width*8+8, y+i/map_width*8, y+i/map_width*8+8, ML_BLACK);
				break;
				case HOLE:
					ML_vertical_line(x+i%map_width*8-1, y+i/map_width*8, y+i/map_width*8+8, ML_BLACK);
					ML_vertical_line(x+i%map_width*8+8, y+i/map_width*8, y+i/map_width*8+8, ML_BLACK);
				break;
				case PLATFORM_LEFT: ML_bmp_8_or_cl(map_bmp[7], x+i%map_width*8, y+i/map_width*8); break;
				case PLATFORM_RIGHT: ML_bmp_8_or_cl(map_bmp[8], x+i%map_width*8, y+i/map_width*8); break;
				case PLATFORM_MIDDLE:
					ML_horizontal_line(y+i/map_width*8, x+i%map_width*8, x+i%map_width*8+7, ML_BLACK);
					ML_horizontal_line(y+i/map_width*8+4, x+i%map_width*8, x+i%map_width*8+7, ML_BLACK);
				break;
				case PLATFORM_LEFT_BOTTOM:
					ML_vertical_line(x+i%map_width*8+6, y+i/map_width*8, y+i/map_width*8+7, ML_BLACK);
				break;
				case PLATFORM_RIGHT_BOTTOM:
					ML_vertical_line(x+i%map_width*8+1, y+i/map_width*8, y+i/map_width*8+7, ML_BLACK);
				break;
				
			}
		}
	}
}

char* int2str(char* c, unsigned int n) {
    if(n==0) {
        c[0] = '0';
        c[1] = 0;
    } else {
        int i, l=0;
        for(i=n ; i ; i/=10)
            l++;
        c[l] = 0;
        for(i=n ; i ; i/=10)
            c[--l] = i%10+'0';
    }
    return c;
}
void game_over() {
	int i;
	PrintXY(35, 20, "Game Over", 0);
	ML_display_vram();
	IsKeyDown(KEY_CTRL_EXE);
	while(i != KEY_CTRL_EXE) GetKey(&i);
}

void pause() {
	int i, j;
	char* vram = ML_vram_adress();
	char bmp[] = {255, 255, 255, 255, 255, 255, 255, 255, 195, 140, 152, 67, 201, 36, 147, 207, 195, 4, 152, 199, 207, 36, 158, 79, 207, 38, 48, 195, 255, 255, 255, 255, 255, 255, 255, 255};
	for(i=0 ; i<9 ; i++)
	{
		for(j=0 ; j<4 ; j++)
			vram[(i+28<<4)+j+6] = bmp[(i<<2)+j];
	}
	IsKeyDown(KEY_CTRL_EXE);
	while(i != KEY_CTRL_EXE) GetKey(&i);
}
int AddIn_main(int isAppli, unsigned short OptionNum) {	
	int i, j, k, l;
	int dx, dy;	
	
	char fps_char[10], points_char[50];
	int fps = 0, fps_frame, fps_last_time;
	
	int gameover = 0;
	int points = 0;	
	
	Hero * hero;
	Alea * alea;
	
	Camera * camera = (Camera *) malloc(sizeof(Camera));
	camera->x = 0;
	camera->y = 0;
	camera->dx = 0;
	camera->dy = 0;
	
	map_x = 0;
	map_y = 0;
	map_dx = MAP_DX;
	map_dy = 0;
	
	alea = (Alea *) malloc(sizeof(Alea));
	alea->type = 1;
	alea->type_nb = 30;
	alea->type_nb_next = 10;
	alea->last_1 = 0;
	alea->last_2 = 0;
	alea->last_3 = 0;
	srand(time_getTicks()+5645673);
	
	fps_last_time = time_getTicks();

	hero = hero_create(hero_bmp);
	
	
	while(1) {
		ML_clear_vram();
		
		/// Update
		
		// Begin
		
		// Hero update
		hero_update(hero);
		if(!hero->anim_isjmp && key_down(K_SHIFT))
			hero_jump_start(hero);
		if(!key_down(K_SHIFT) && hero->dy < 0 && !hero->anim_isfall)
			hero_jump_reset(hero);
		
		// Hero colision
		if(!gameover) {
			j = (hero->y+7-map_y)/8; // Ligne dans la matrice
			i = (hero->x-map_x)/8+j*18; // Collision avec arriere
			k = (hero->x+3-map_x)/8+j*18; // Collision equilibr
			l = (hero->x+8-map_x)/8+j*18; // Collision avec l'avant
			// ML_rectangle(camera->x+l%map_width*8, camera->y+l/map_width*8, camera->x+l%map_width*8+8, camera->y+l/map_width*8+8, 1, ML_BLACK, ML_TRANSPARENT);		
			if(i > 0 && (map[l+18] >= SPRITE_SOLID_MIN && map[l+18] <= SPRITE_SOLID_MAX && map[l+18] != PICK) && hero->anim_isjmp && hero->dy > 0) { // sol, block ou boite
				hero_jump_stop(hero, map_y+l/map_width*8-4);
			} else if(i > 0 && (map[i+1+18] == PICK) && hero->anim_isjmp && hero->dy > 0) { // sur pick
				map_dx = 0;
				gameover = 1;
			} else if(i > 0 && map[i+1] >= SPRITE_SOLID_MIN && map[i+1] <= SPRITE_SOLID_MAX) { // cogner contre nimporte quoi
				map_dx = 0;
				gameover = 1;
				hero->anim_bmp = 3; // Jump anim
			} else if(i > 0 && map[k+18] == HOLE && !hero->anim_isjmp) { // dans un trou (1 case)
				map_dx = 0;
				gameover = 1;
				hero_falling_start(hero);
			} else if(i > 0 && map[i+18] >= HOLE_LEFT && map[i+18] <= HOLE_RIGHT && !hero->anim_isjmp) { // dans un trou (2 cases)
				if(map[l+18] >= HOLE_LEFT && map[l+18] <= HOLE_RIGHT) {					
					map_dx = 0;
					gameover = 1;
					hero_falling_start(hero);
				}		
			} else if(i > 0 && map[i+18] == NONE && !hero->anim_isjmp) { // vide en dessous
				hero_falling_start(hero);
			}
			
			if(hero->y > 63) {
				gameover = 1;
				map_dx = 0;
			}
			
			/// God mode :)
			// gameover = 0;
			// map_dx = MAP_DX;
		}		
		
		// Map
		if(time_getTicks() - map_last_update > 2) {
			map_x -= map_dx;
			
			map_last_update = time_getTicks();
			
			// Score
			if(!gameover) {
				points += 3;
			}
			
			if(map_x <= -8) { // Time to gen
				map_x = 0;
				map_gen(map, map_width, map_height, alea); // gen next column			
			}
		}
		
		// Camera			
		dy = hero->y-36+camera->y;
		if(hero->y < 36) {
			dy = hero->y-26+camera->y;
		}		
		
		if((dy < 0 && !hero->anim_isjmp) || dy < -15) {
			camera->y += CAMERA_SPEED_Y; // monte
		} else if ((dy > 0 && !hero->anim_isjmp) || (dy < 0 && camera->y > 0 && !hero->anim_isjmp) || dy > 15) {
			camera->y -= CAMERA_SPEED_Y; // inv
		}		
		
		// Fps
		if(time_getTicks()-fps_last_time>128) { 
			fps = fps_frame;
			if(fps < 0 || fps > 1000) fps = 0;
			fps_frame = 0;
			fps_last_time = time_getTicks();
			
			if(!gameover && map_dx <= 3) // Up the difficulty par second
				map_dx += 0.03;
		}
		fps_frame++;
		
		// End Update		
		
		/// Draw
		
		// Begin
		
		// Hero
		ML_bmp_or_cl(hero->bmp[hero->anim_bmp], hero->x+(int) (camera->x), hero->y+(int) (camera->y), 10, 12);		
		
		// Map
		map_draw(camera);		
		
		// Hud
		int2str(fps_char, fps);
		// int2str(fps_char, (int) map_dx);		
		PrintMini(127-15, 1, fps_char, 0);
		
		int2str(points_char, (int) points);
		PrintMini(1, 1, points_char, 0);
		
		// End		
		
		if(gameover) { // Game over
			hero_jump_start(hero);
			hero->anim_isfall = 1;
			if(hero->y > 64) {
				game_over();
				camera->x = 0;
				camera->y = 0;
				camera->dx = 0;
				camera->dy = 0;
				
				map_x = 0;
				map_y = 0;
				map_dx = MAP_DX;
				map_dy = 0;
				
				alea->type = 1;
				alea->type_nb = 30;
				alea->type_nb_next = 10;
				alea->last_1 = 0;
				alea->last_2 = 0;
				alea->last_3 = 0;
				
				points = 0;
				
				memcpy(map, map_backup, map_width*map_height*sizeof(unsigned char));
				
				hero_reset(hero);
				
				gameover = 0;
			}
		}	
		
		if(key_down(K_EXIT))
			pause();
		
		ML_display_vram();
		
		Sleep(10);
    }

    return 1;
}

#pragma section _BR_Size
unsigned long BR_Size;
#pragma section

#pragma section _TOP

int InitializeSystem(int isAppli, unsigned short OptionNum)
{
    return INIT_ADDIN_APPLICATION(isAppli, OptionNum);
}

#pragma section

